#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Unsupported contrib code for OpenTimelineIO."""

# flake8: noqa

from . import (
    adapters
)
